"""Tests for the searchspace module."""

import numpy as np
import pandas as pd
import pytest
from pandas.testing import assert_frame_equal

from baybe._optional.info import POLARS_INSTALLED
from baybe.constraints import (
    ContinuousCardinalityConstraint,
    ContinuousLinearConstraint,
    DiscreteSumConstraint,
    ThresholdCondition,
)
from baybe.exceptions import EmptySearchSpaceError
from baybe.parameters import (
    CategoricalParameter,
    NumericalContinuousParameter,
    NumericalDiscreteParameter,
)
from baybe.searchspace import (
    SearchSpace,
    SearchSpaceType,
    SubspaceContinuous,
    SubspaceDiscrete,
)
from baybe.searchspace.discrete import (
    parameter_cartesian_prod_pandas,
    parameter_cartesian_prod_polars,
)


def test_empty_parameters():
    """Creation of a search space with no parameters raises an exception."""
    with pytest.raises(EmptySearchSpaceError):
        SearchSpace()


def test_bounds_order():
    """Asserts that the bounds are created in the correct order.

    The correct order is discrete parameters first, continuous next.
    """
    parameters = [
        NumericalDiscreteParameter(name="A_disc", values=[1.0, 2.0, 3.0]),
        NumericalContinuousParameter(name="A_cont", bounds=(4.0, 6.0)),
        NumericalDiscreteParameter(name="B_disc", values=[7.0, 8.0, 9.0]),
        NumericalContinuousParameter(name="B_cont", bounds=(10.0, 12.0)),
    ]
    searchspace = SearchSpace.from_product(parameters=parameters)
    expected = np.array([[1.0, 7.0, 4.0, 10.0], [3.0, 9.0, 6.0, 12.0]])
    assert np.array_equal(
        searchspace.comp_rep_bounds.values,
        expected,
    )


def test_empty_parameter_bounds():
    """Asserts that the correct bounds are produced for empty search spaces.

    Also checks for the correct shapes.
    """
    parameters = []
    searchspace_discrete = SubspaceDiscrete.from_product(parameters=parameters)
    searchspace_continuous = SubspaceContinuous(parameters=parameters)
    expected = pd.DataFrame(np.empty((2, 0)), index=["min", "max"])
    pd.testing.assert_frame_equal(searchspace_discrete.comp_rep_bounds, expected)
    pd.testing.assert_frame_equal(searchspace_continuous.comp_rep_bounds, expected)


def test_discrete_searchspace_creation_from_dataframe():
    """A purely discrete search space is created from an example dataframe."""
    num_specified = NumericalDiscreteParameter(name="num_specified", values=[1, 2, 3])
    num_unspecified = NumericalDiscreteParameter(
        name="num_unspecified", values=[4, 5, 6]
    )
    cat_specified = CategoricalParameter(name="cat_specified", values=["a", "b", "c"])
    cat_unspecified = CategoricalParameter(
        name="cat_unspecified", values=["d", "e", "f"]
    )

    all_params = (cat_specified, cat_unspecified, num_specified, num_unspecified)

    df = pd.DataFrame({param.name: param.values for param in all_params})
    searchspace = SearchSpace(
        SubspaceDiscrete.from_dataframe(df, parameters=[num_specified, cat_specified])
    )

    assert searchspace.type == SearchSpaceType.DISCRETE
    assert searchspace.parameters == all_params
    assert df.equals(searchspace.discrete.exp_rep)


def test_discrete_from_dataframe_dtype_consistency():
    """Inconsistent but valid dtypes are correctly converted."""
    df = pd.DataFrame(
        {
            "A": [1, 2, 3],
            "B": ["x", "y", "z"],
            "C": [int(1), 2.2, True],  # Valid but inconsistent dtypes
        }
    )
    subspace = SubspaceDiscrete.from_dataframe(df)

    assert isinstance(
        next(p for p in subspace.parameters if p.name == "C"),
        NumericalDiscreteParameter,
    )
    assert pd.api.types.is_float_dtype(subspace.exp_rep["C"])


def test_invalid_simplex_creating_with_overlapping_parameters():
    """Creating a simplex searchspace with overlapping simplex and product parameters
    raises an error."""  # noqa
    parameters = [NumericalDiscreteParameter(name="x_1", values=(0, 1, 2))]

    with pytest.raises(
        ValueError,
        match="'simplex_parameters' and 'product_parameters' must be disjoint",
    ):
        SearchSpace(
            SubspaceDiscrete.from_simplex(
                max_sum=1.0,
                simplex_parameters=parameters,
                product_parameters=parameters,
            )
        )


def test_continuous_searchspace_creation_from_bounds():
    """A purely continuous search space is created from example bounds."""
    parameters = (
        NumericalContinuousParameter("param1", (0, 1)),
        NumericalContinuousParameter("param2", (-1, 1)),
    )
    bounds = pd.DataFrame({p.name: p.bounds.to_tuple() for p in parameters})
    searchspace = SearchSpace(continuous=SubspaceContinuous.from_bounds(bounds))

    assert searchspace.type == SearchSpaceType.CONTINUOUS
    assert searchspace.parameters == parameters


def test_hyperrectangle_searchspace_creation():
    """A purely continuous search space is created that spans a certain set of points.

    As the name suggests, this searchspace is hyperrectangle- shaped
    """
    points = pd.DataFrame(
        {
            "param1": [0, 1, 2],
            "param2": [-1, 0, 1],
        }
    )
    searchspace = SearchSpace(continuous=SubspaceContinuous.from_dataframe(points))

    parameters = (
        NumericalContinuousParameter("param1", (0, 2)),
        NumericalContinuousParameter("param2", (-1, 1)),
    )

    assert searchspace.type == SearchSpaceType.CONTINUOUS
    assert searchspace.parameters == parameters


def test_invalid_constraint_parameter_combos():
    """Testing invalid constraint-parameter combinations."""
    parameters = [
        CategoricalParameter("cat1", values=("c1", "c2")),
        NumericalDiscreteParameter("d1", values=[1, 2, 3]),
        NumericalDiscreteParameter("d2", values=[0, 1, 2]),
        NumericalContinuousParameter("c1", (0, 2)),
        NumericalContinuousParameter("c2", (-1, 1)),
    ]

    # Attempting continuous constraint over hybrid parameter set
    with pytest.raises(ValueError):
        SearchSpace.from_product(
            parameters=parameters,
            constraints=[ContinuousLinearConstraint(["c1", "c2", "d1"], "=")],
        )

    # Attempting continuous constraint over hybrid parameter set
    with pytest.raises(ValueError):
        SearchSpace.from_product(
            parameters=parameters,
            constraints=[ContinuousLinearConstraint(["c1", "c2", "d1"], "=")],
        )

    # Attempting discrete constraint over hybrid parameter set
    with pytest.raises(ValueError):
        SearchSpace.from_product(
            parameters=parameters,
            constraints=[
                DiscreteSumConstraint(
                    parameters=["d1", "d2", "c1"],
                    condition=ThresholdCondition(threshold=1.0, operator=">"),
                )
            ],
        )

    # Attempting constraints over parameter set where a parameter does not exist
    with pytest.raises(ValueError):
        SearchSpace.from_product(
            parameters=parameters,
            constraints=[
                DiscreteSumConstraint(
                    parameters=["d1", "e7", "c1"],
                    condition=ThresholdCondition(threshold=1.0, operator=">"),
                )
            ],
        )

    # Attempting constraints over parameter set where a parameter does not exist
    with pytest.raises(ValueError):
        SearchSpace.from_product(
            parameters=parameters,
            constraints=[ContinuousLinearConstraint(["c1", "e7", "d1"], "=")],
        )

    # Attempting constraints over parameter sets containing non-numerical discrete
    # parameters.
    with pytest.raises(
        ValueError, match="valid only for numerical discrete parameters"
    ):
        SearchSpace.from_product(
            parameters=parameters,
            constraints=[
                DiscreteSumConstraint(
                    parameters=["cat1", "d1", "d2"],
                    condition=ThresholdCondition(threshold=1.0, operator=">"),
                )
            ],
        )


@pytest.mark.parametrize(
    "parameter_names",
    [
        [
            "Categorical_1",
            "Categorical_2",
            "Frame_A",
            "Some_Setting",
            "Num_disc_1",
            "Fraction_1",
            "Solvent_1",
            "Custom_1",
        ],
        [
            "Categorical_1_subset",
            "Categorical_2",
            "Frame_A",
            "Some_Setting",
            "Num_disc_1",
            "Fraction_1",
            "Solvent_1_subset",
            "Custom_1_subset",
            "Task",
        ],
    ],
    ids=["simple", "with_active_values"],
)
def test_searchspace_memory_estimate(searchspace: SearchSpace):
    """The memory estimate doesn't differ by more than 5% from the actual memory."""
    estimate = searchspace.estimate_product_space_size(searchspace.parameters)
    estimate_exp = estimate.exp_rep_bytes
    estimate_comp = estimate.comp_rep_bytes

    actual_exp = searchspace.discrete.exp_rep.memory_usage(deep=True, index=False).sum()
    actual_comp = searchspace.discrete.comp_rep.memory_usage(
        deep=True, index=False
    ).sum()

    assert 0.95 <= estimate_exp / actual_exp <= 1.05, (
        "Exp: ",
        estimate_exp,
        actual_exp,
    )
    assert 0.95 <= estimate_comp / actual_comp <= 1.05, (
        "Comp: ",
        estimate_comp,
        actual_comp,
    )


def test_cardinality_constraints_with_overlapping_parameters():
    """Creating cardinality constraints with overlapping parameters raises an error."""
    parameters = (
        NumericalContinuousParameter("c1", (0, 1)),
        NumericalContinuousParameter("c2", (0, 1)),
        NumericalContinuousParameter("c3", (0, 1)),
    )
    with pytest.raises(ValueError, match="cannot share the same parameters"):
        SubspaceContinuous(
            parameters=parameters,
            constraints_nonlin=(
                ContinuousCardinalityConstraint(
                    parameters=["c1", "c2"],
                    max_cardinality=1,
                ),
                ContinuousCardinalityConstraint(
                    parameters=["c2", "c3"],
                    max_cardinality=1,
                ),
            ),
        )


def test_cardinality_constraint_with_invalid_parameter_bounds():
    """Imposing a cardinality constraint on a parameter whose range does not include
    zero raises an error."""  # noqa
    parameters = (
        NumericalContinuousParameter("c1", (0, 1)),
        NumericalContinuousParameter("c2", (1, 2)),
    )
    with pytest.raises(ValueError, match="must include zero"):
        SubspaceContinuous(
            parameters=parameters,
            constraints_nonlin=(
                ContinuousCardinalityConstraint(
                    parameters=["c1", "c2"],
                    max_cardinality=1,
                ),
            ),
        )


@pytest.mark.skipif(
    not POLARS_INSTALLED, reason="Optional polars dependency not installed."
)
@pytest.mark.parametrize(
    "parameter_names",
    [
        [
            "Categorical_1",
            "Categorical_2",
            "Custom_1",
            "Solvent_1",
            "Solvent_2",
            "Fraction_1",
        ],
        [
            "Categorical_1_subset",
            "Categorical_2",
            "Custom_1_subset",
            "Solvent_1_subset",
            "Solvent_2",
            "Fraction_1",
        ],
    ],
    ids=["simple", "with_active_values"],
)
def test_polars_pandas_equivalence(parameters):
    """Search spaces created with Polars and Pandas are identical."""
    # Do Polars product
    ldf = parameter_cartesian_prod_polars(parameters)
    df_pl = ldf.collect()

    # Do Pandas product
    df_pd = parameter_cartesian_prod_pandas(parameters)

    # Assert equality
    assert_frame_equal(df_pl.to_pandas(), df_pd)



"""Surrogate tests."""

from unittest.mock import patch

import pandas as pd
import pytest
from pytest import param

from baybe.objectives.pareto import ParetoObjective
from baybe.parameters.numerical import NumericalDiscreteParameter
from baybe.recommenders.pure.bayesian.botorch import BotorchRecommender
from baybe.surrogates import BayesianLinearSurrogate, NGBoostSurrogate
from baybe.surrogates.composite import CompositeSurrogate
from baybe.surrogates.gaussian_process.core import GaussianProcessSurrogate
from baybe.surrogates.random_forest import RandomForestSurrogate
from baybe.targets.numerical import NumericalTarget


@patch.object(GaussianProcessSurrogate, "_fit")
def test_caching(patched, searchspace, objective, fake_measurements):
    """A second fit call with the same context does not trigger retraining."""
    # Prepare the setting
    surrogate = GaussianProcessSurrogate()

    # First call
    surrogate.fit(searchspace, objective, fake_measurements)
    patched.assert_called()

    patched.reset_mock()

    # Second call
    surrogate.fit(searchspace, objective, fake_measurements)
    patched.assert_not_called()


@pytest.mark.parametrize(
    "surrogate",
    [
        CompositeSurrogate(
            {"t1": GaussianProcessSurrogate(), "t2": RandomForestSurrogate()},
        ),
        CompositeSurrogate.from_replication(GaussianProcessSurrogate()),
    ],
    ids=["via_init", "via_template"],
)
def test_composite_surrogates(surrogate):
    """Composition yields a valid surrogate."""
    t1 = NumericalTarget("t1", "MAX")
    t2 = NumericalTarget("t2", "MIN")
    searchspace = NumericalDiscreteParameter("p", [0, 1]).to_searchspace()
    objective = ParetoObjective([t1, t2])
    measurements = pd.DataFrame({"p": [0], "t1": [0], "t2": [0]})
    BotorchRecommender(surrogate_model=surrogate).recommend(
        2, searchspace, objective, measurements
    )


@pytest.mark.parametrize(
    "model_cls, params",
    [
        param(RandomForestSurrogate, {"n_estimators": 5}, id="rf_1"),
        param(RandomForestSurrogate, {"criterion": "squared_error"}, id="rf_2"),
        param(RandomForestSurrogate, {"max_features": 0.4}, id="rf_3"),
        param(RandomForestSurrogate, {"monotonic_cst": None}, id="rf_4"),
        param(NGBoostSurrogate, {"n_estimators": 5}, id="ng_1"),
        param(NGBoostSurrogate, {"learning_rate": 0.05}, id="ng_2"),
        param(NGBoostSurrogate, {"verbose": True}, id="ng_3"),
        param(NGBoostSurrogate, {"early_stopping_rounds": None}, id="ng_4"),
        param(BayesianLinearSurrogate, {"max_iter": 10}, id="lin_1"),
        param(BayesianLinearSurrogate, {"alpha_1": 0.3}, id="lin_2"),
        param(BayesianLinearSurrogate, {"fit_intercept": True}, id="lin_3"),
    ],
)
def test_valid_model_params(model_cls, params):
    """Ensure valid model_params can be set."""
    model_cls(model_params=params)


@pytest.mark.parametrize(
    "model_cls, params",
    [
        param(RandomForestSurrogate, {"n_estimators": 5.3}, id="rf_float_not_int"),
        param(
            RandomForestSurrogate,
            {"criterion": "squared_error123"},
            id="rf_wrong_literal",
        ),
        param(RandomForestSurrogate, {"max_features": True}, id="rf_unsupported_bool"),
        param(
            RandomForestSurrogate, {"min_samples_split": None}, id="rf_unsupported_none"
        ),
        param(RandomForestSurrogate, {"bla": None}, id="rf_extra_key"),
        param(NGBoostSurrogate, {"n_estimators": 5.3}, id="ng_float_not_int"),
        param(NGBoostSurrogate, {"learning_rate": 4}, id="ng_int_not_float"),
        param(NGBoostSurrogate, {"verbose": 1.0}, id="ng_float_not_bool"),
        param(
            NGBoostSurrogate, {"early_stopping_rounds": 3.4}, id="ng_unsupported_float"
        ),
        param(NGBoostSurrogate, {"col_sample": None}, id="ng_unsupported_none"),
        param(NGBoostSurrogate, {"bla": 3.4}, id="ng_extra_key"),
        param(BayesianLinearSurrogate, {"max_iter": 10.3}, id="lin_float_not_int"),
        param(BayesianLinearSurrogate, {"alpha_1": 4}, id="lin_int_not_float"),
        param(BayesianLinearSurrogate, {"fit_intercept": 0.0}, id="lin_float_not_bool"),
        param(BayesianLinearSurrogate, {"copy_X": None}, id="lin_unsupported_none"),
        param(BayesianLinearSurrogate, {"bla": 0.0}, id="lin_extra_key"),
    ],
)
def test_invalid_model_params(model_cls, params):
    """Ensure valid model_params cannot be set."""
    with pytest.raises(TypeError, match="model_params"):
        model_cls(model_params=params)



"""Tests for the continuous cardinality constraint."""

import warnings
from collections.abc import Sequence
from itertools import combinations_with_replacement
from warnings import WarningMessage

import numpy as np
import pandas as pd
import pytest

from baybe.constraints.continuous import (
    ContinuousCardinalityConstraint,
    ContinuousLinearConstraint,
)
from baybe.constraints.utils import is_cardinality_fulfilled
from baybe.exceptions import MinimumCardinalityViolatedWarning
from baybe.parameters.numerical import NumericalContinuousParameter
from baybe.recommenders import BotorchRecommender
from baybe.recommenders.pure.bayesian.base import BayesianRecommender
from baybe.recommenders.pure.nonpredictive.sampling import RandomRecommender
from baybe.searchspace import SearchSpace, SubspaceContinuous
from baybe.targets.numerical import NumericalTarget


def _validate_cardinality_constrained_batch(
    batch: pd.DataFrame,
    subspace_continuous: SubspaceContinuous,
    batch_size: int,
    captured_warnings: Sequence[WarningMessage],
):
    """Validate that a cardinality-constrained batch fulfills the necessary conditions.

    Args:
        batch: The batch to validate.
        subspace_continuous: The continuous subspace from which to recommend the points.
        batch_size: The number of points to be recommended.
        captured_warnings: A list of captured warnings.
    """
    # Assert that the maximum cardinality constraint is fulfilled
    assert is_cardinality_fulfilled(batch, subspace_continuous, check_minimum=False)

    # Check whether the minimum cardinality constraint is fulfilled
    is_min_cardinality_fulfilled = is_cardinality_fulfilled(
        batch, subspace_continuous, check_maximum=False
    )

    # A warning must be raised when the minimum cardinality constraint is not fulfilled
    cardinality_warnings = [
        w
        for w in captured_warnings
        if issubclass(w.category, MinimumCardinalityViolatedWarning)
    ]
    assert is_min_cardinality_fulfilled != bool(cardinality_warnings)

    # Assert that we obtain as many samples as requested
    assert batch.shape[0] == batch_size

    # Sanity check: If all recommendations in the batch are identical, something is
    # fishy – unless the cardinality is 0, in which case the entire batch must contain
    # zeros. Technically, the probability of getting such a degenerate batch
    # is not zero, hence this is not a strict requirement. However, in earlier BoTorch
    # versions, this simply happened due to a bug in their sampler:
    # https://github.com/pytorch/botorch/issues/2351
    # We thus include this check as a safety net for catching regressions. If it
    # turns out the check fails because we observe degenerate batches as actual
    # recommendations, we need to invent something smarter.
    max_cardinalities = [
        c.max_cardinality for c in subspace_continuous.constraints_cardinality
    ]
    if len(unique_row := batch.drop_duplicates()) == 1:
        assert (unique_row.iloc[0] == 0.0).all() and all(
            max_cardinality == 0 for max_cardinality in max_cardinalities
        )


# Combinations of cardinalities to be tested
cardinality_bounds_combinations = sorted(combinations_with_replacement(range(0, 10), 2))


@pytest.mark.parametrize(
    "cardinality_bounds",
    cardinality_bounds_combinations,
    ids=[str(x) for x in cardinality_bounds_combinations],
)
def test_sampling_cardinality_constraint(cardinality_bounds: tuple[int, int]):
    """Sampling on unit-cube with cardinality constraints respects all constraints and
    produces distinct samples."""  # noqa

    N_PARAMETERS = 10
    BATCH_SIZE = 10
    min_cardinality, max_cardinality = cardinality_bounds

    parameters = tuple(
        NumericalContinuousParameter(name=f"x_{i}", bounds=(0, 1))
        for i in range(N_PARAMETERS)
    )

    constraints = (
        ContinuousCardinalityConstraint(
            parameters=[f"x_{i}" for i in range(N_PARAMETERS)],
            min_cardinality=min_cardinality,
            max_cardinality=max_cardinality,
        ),
    )

    subspace_continous = SubspaceContinuous(
        parameters=parameters, constraints_nonlin=constraints
    )

    with warnings.catch_warnings(record=True) as w:
        samples = subspace_continous.sample_uniform(BATCH_SIZE)

    # Assert that the constraint conditions hold
    _validate_cardinality_constrained_batch(samples, subspace_continous, BATCH_SIZE, w)


def test_polytope_sampling_with_cardinality_constraint():
    """Polytope sampling with cardinality constraints respects all involved
    constraints and produces distinct samples."""  # noqa

    N_PARAMETERS = 6
    MAX_CARDINALITY = 4
    MIN_CARDINALITY = 2
    BATCH_SIZE = 20
    TOLERANCE = 1e-3

    parameters = [
        NumericalContinuousParameter(name=f"x_{i + 1}", bounds=(0, 1))
        for i in range(N_PARAMETERS)
    ]
    params_equality = ["x_1", "x_2", "x_3", "x_4"]
    coeffs_equality = [0.9, 0.6, 2.8, 6.1]
    rhs_equality = 4.2
    params_inequality = ["x_1", "x_2", "x_5", "x_6"]
    coeffs_inequality = [4.7, 1.4, 4.6, 8.6]
    rhs_inequality = 1.3
    params_cardinality = ["x_1", "x_2", "x_3", "x_5"]
    constraints = [
        ContinuousLinearConstraint(
            parameters=params_equality,
            operator="=",
            coefficients=coeffs_equality,
            rhs=rhs_equality,
        ),
        ContinuousLinearConstraint(
            parameters=params_inequality,
            operator=">=",
            coefficients=coeffs_inequality,
            rhs=rhs_equality,
        ),
        ContinuousCardinalityConstraint(
            parameters=params_cardinality,
            max_cardinality=MAX_CARDINALITY,
            min_cardinality=MIN_CARDINALITY,
        ),
    ]
    subspace_continous = SubspaceContinuous.from_product(parameters, constraints)

    with warnings.catch_warnings(record=True) as w:
        samples = subspace_continous.sample_uniform(BATCH_SIZE)

    # Assert that the constraint conditions hold
    _validate_cardinality_constrained_batch(samples, subspace_continous, BATCH_SIZE, w)

    # Assert that linear equality constraint is fulfilled
    assert np.allclose(
        np.sum(samples[params_equality] * coeffs_equality, axis=1),
        rhs_equality,
        atol=TOLERANCE,
    )

    # Assert that linear non-equality constraint is fulfilled
    assert (
        np.sum(samples[params_inequality] * coeffs_inequality, axis=1)
        .ge(rhs_inequality - TOLERANCE)
        .all()
    )


def test_min_cardinality_warning():
    """Providing candidates violating minimum cardinality constraint raises a
    warning.
    """  # noqa
    N_PARAMETERS = 2
    MIN_CARDINALITY = 2
    MAX_CARDINALITY = 2
    BATCH_SIZE = 20

    lower_bound = -0.5
    upper_bound = 0.5
    stepsize = 0.05
    parameters = [
        NumericalContinuousParameter(
            name=f"x_{i + 1}", bounds=(lower_bound, upper_bound)
        )
        for i in range(N_PARAMETERS)
    ]

    constraints = [
        ContinuousCardinalityConstraint(
            parameters=[p.name for p in parameters],
            max_cardinality=MAX_CARDINALITY,
            min_cardinality=MIN_CARDINALITY,
        ),
    ]

    searchspace = SearchSpace.from_product(parameters, constraints)
    objective = NumericalTarget("t", "MAX").to_objective()

    # Create a scenario in which
    # - The optimum of the target function is at the origin
    # - The Botorch recommender is likely to provide candidates at the origin,
    # which violates the minimum cardinality constraint.
    def custom_target(x1: np.ndarray, x2: np.ndarray) -> np.ndarray:
        """A custom target function with maximum at the origin."""
        return -abs(x1) - abs(x2)

    def prepare_measurements() -> pd.DataFrame:
        """Prepare measurements."""
        x1 = np.arange(lower_bound, upper_bound + stepsize, stepsize)
        # Exclude 0 from the array
        X1, X2 = np.meshgrid(x1[abs(x1) > stepsize / 2], x1[abs(x1) > stepsize / 2])

        return pd.DataFrame(
            {
                "x_1": X1.flatten(),
                "x_2": X2.flatten(),
                "t": custom_target(X1.flatten(), X2.flatten()),
            }
        )

    with warnings.catch_warnings(record=True) as captured_warnings:
        BotorchRecommender().recommend(
            BATCH_SIZE, searchspace, objective, prepare_measurements()
        )
    assert any(
        issubclass(w.category, MinimumCardinalityViolatedWarning)
        for w in captured_warnings
    )


def test_empty_constraints_after_cardinality_constraint():
    """Constraints that have no more parameters left due to activated
    cardinality constraints do not cause crashes."""  # noqa

    N_PARAMETERS = 2

    parameters = [
        NumericalContinuousParameter(name=f"x_{i + 1}", bounds=(0, 1))
        for i in range(N_PARAMETERS)
    ]
    constraints = [
        ContinuousLinearConstraint(
            parameters=["x_1"],
            operator="=",
            coefficients=[1.0],
            rhs=0.3,
        ),
        ContinuousLinearConstraint(
            parameters=["x_2"],
            operator="<=",
            coefficients=[1.0],
            rhs=0.6,
        ),
        ContinuousCardinalityConstraint(
            parameters=["x_1", "x_2"],
            max_cardinality=1,
            min_cardinality=1,
        ),
    ]
    subspace = SubspaceContinuous.from_product(parameters, constraints)
    subspace.sample_uniform(1)


@pytest.mark.parametrize("recommender", [RandomRecommender(), BotorchRecommender()])
def test_cardinality_constraint(recommender):
    """Cardinality constraints are taken into account by the recommender."""
    MIN_CARDINALITY = 4
    MAX_CARDINALITY = 7
    BATCH_SIZE = 10

    parameters = [NumericalContinuousParameter(str(i), (0, 1)) for i in range(10)]
    constraints = [
        ContinuousCardinalityConstraint(
            [p.name for p in parameters], MIN_CARDINALITY, MAX_CARDINALITY
        )
    ]
    searchspace = SearchSpace.from_product(parameters, constraints)

    if isinstance(recommender, BayesianRecommender):
        objective = NumericalTarget("t", "MAX").to_objective()
        measurements = pd.DataFrame(searchspace.continuous.sample_uniform(2))
        measurements["t"] = np.random.random(len(measurements))
    else:
        objective = None
        measurements = None

    with warnings.catch_warnings(record=True) as w:
        recommendation = recommender.recommend(
            BATCH_SIZE, searchspace, objective, measurements
        )

    # Assert that the constraint conditions hold
    _validate_cardinality_constrained_batch(
        recommendation, searchspace.continuous, BATCH_SIZE, w
    )



"""Benchmark definitions."""

from benchmarks.definition.base import (
    Benchmark,
    BenchmarkSettings,
)
from benchmarks.definition.convergence import (
    ConvergenceBenchmark,
    ConvergenceBenchmarkSettings,
)

__all__ = [
    "Benchmark",
    "BenchmarkSettings",
    "ConvergenceBenchmark",
    "ConvergenceBenchmarkSettings",
]



"""Linear surrogates."""

from __future__ import annotations

import gc
from typing import TYPE_CHECKING, ClassVar, TypedDict

from attrs import define, field
from typing_extensions import override

from baybe.surrogates.base import IndependentGaussianSurrogate
from baybe.surrogates.utils import batchify_mean_var_prediction, catch_constant_targets
from baybe.surrogates.validation import make_dict_validator
from baybe.utils.conversion import to_string

if TYPE_CHECKING:
    from torch import Tensor


class _ARDRegressionParams(TypedDict, total=False):
    """Optional ARDRegression parameters.

    See :class:`~sklearn.linear_model.ARDRegression`.
    """

    max_iter: int
    tol: float
    alpha_1: float
    alpha_2: float
    lambda_1: float
    lambda_2: float
    compute_score: bool
    threshold_lambda: float
    fit_intercept: bool
    copy_X: bool
    verbose: bool


@catch_constant_targets
@define
class BayesianLinearSurrogate(IndependentGaussianSurrogate):
    """A Bayesian linear regression surrogate model."""

    supports_transfer_learning: ClassVar[bool] = False
    # See base class.

    model_params: _ARDRegressionParams = field(
        factory=dict,
        converter=dict,
        validator=make_dict_validator(_ARDRegressionParams),
    )
    """Optional model parameter that will be passed to the surrogate constructor.

    For allowed keys and values, see :class:`~sklearn.linear_model.ARDRegression`.
    """

    # TODO: type should be `ARDRegression | None` but is currently omitted due to:
    #  https://github.com/python-attrs/cattrs/issues/531
    _model = field(init=False, default=None, eq=False)
    """The actual model."""

    @override
    @batchify_mean_var_prediction
    def _estimate_moments(
        self, candidates_comp_scaled: Tensor, /
    ) -> tuple[Tensor, Tensor]:
        # FIXME[typing]: It seems there is currently no better way to inform the type
        #   checker that the attribute is available at the time of the function call
        assert self._model is not None

        import torch

        # Get predictions
        dists = self._model.predict(candidates_comp_scaled.numpy(), return_std=True)

        # Split into posterior mean and variance
        mean = torch.from_numpy(dists[0])
        var = torch.from_numpy(dists[1]).pow(2)

        return mean, var

    @override
    def _fit(self, train_x: Tensor, train_y: Tensor) -> None:
        from sklearn.linear_model import ARDRegression

        self._model = ARDRegression(**(self.model_params))
        self._model.fit(train_x, train_y.ravel())

    @override
    def __str__(self) -> str:
        fields = [to_string("Model Params", self.model_params, single_line=True)]
        return to_string(super().__str__(), *fields)


# Collect leftover original slotted classes processed by `attrs.define`
gc.collect()



"""Test serialization of constraints."""

import hypothesis.strategies as st
import pytest
from hypothesis import given
from pytest import param

from baybe.constraints.base import Constraint

from ..hypothesis_strategies.constraints import (
    continuous_linear_constraints,
    discrete_dependencies_constraints,
    discrete_excludes_constraints,
    discrete_linked_parameters_constraints,
    discrete_no_label_duplicates_constraints,
    discrete_permutation_invariance_constraints,
    discrete_product_constraints,
    discrete_sum_constraints,
)


@pytest.mark.parametrize(
    "constraint_strategy",
    [
        param(
            discrete_permutation_invariance_constraints(),
            id="DiscretePermutationInvarianceConstraint",
        ),
        param(discrete_dependencies_constraints(), id="DiscreteDependenciesConstraint"),
        param(discrete_excludes_constraints(), id="DiscreteExcludeConstraint"),
        param(discrete_sum_constraints(), id="DiscreteSumConstraint"),
        param(discrete_product_constraints(), id="DiscreteProductConstraint"),
        param(
            discrete_no_label_duplicates_constraints(),
            id="DiscreteNoLabelDuplicatesConstraint",
        ),
        param(
            discrete_linked_parameters_constraints(),
            id="DiscreteLinkedParametersConstraint",
        ),
        param(
            continuous_linear_constraints(),
            id="ContinuousLinearonstraint",
        ),
    ],
)
@given(data=st.data())
def test_constraint_roundtrip(constraint_strategy, data):
    """A serialization roundtrip yields an equivalent object."""
    constraint = data.draw(constraint_strategy)
    string = constraint.to_json()
    constraint2 = Constraint.from_json(string)
    assert constraint == constraint2, (constraint, constraint2)


@pytest.mark.parametrize(
    "constraint_names",
    [["Constraint_13"]],
)
@pytest.mark.parametrize("n_grid_points", [5])
def test_unsupported_serialization_of_custom_constraint(constraints):
    with pytest.raises(NotImplementedError):
        constraint = constraints[0]
        constraint.to_json()



"""Validation tests for parameters."""

from unittest.mock import Mock

import numpy as np
import pandas as pd
import pytest
from cattrs.errors import IterableValidationError
from pytest import param

from baybe._optional.info import CHEM_INSTALLED
from baybe.parameters.categorical import (
    CategoricalParameter,
    TaskParameter,
)
from baybe.parameters.custom import CustomDiscreteParameter
from baybe.parameters.numerical import (
    NumericalContinuousParameter,
    NumericalDiscreteParameter,
)
from baybe.parameters.substance import SubstanceParameter
from baybe.parameters.validation import validate_decorrelation
from baybe.utils.interval import InfiniteIntervalError

try:  # For python < 3.11, use the exceptiongroup backport
    ExceptionGroup
except NameError:
    from exceptiongroup import ExceptionGroup


@pytest.mark.parametrize(
    ("name", "error"),
    [
        param("", ValueError, id="empty_string"),
        param(1, TypeError, id="not_a_string"),
    ],
)
def test_invalid_parameter_name(name, error):
    """Providing an invalid parameter name raises an exception."""
    with pytest.raises(error):
        NumericalDiscreteParameter(name=name, values=[1, 2, 3])


@pytest.mark.parametrize(
    ("decorrelation", "error"),
    [
        param(1, TypeError, id="wrong_type"),
        param(1.0, ValueError, id="too_high"),
        param(0.0, ValueError, id="too_low"),
        param(float("nan"), ValueError, id="nan"),
        param(float("inf"), ValueError, id="inf"),
    ],
)
def test_invalid_decorrelation(decorrelation, error):
    """Providing an invalid decorrelation setting raises an exception."""
    with pytest.raises(error):
        validate_decorrelation(Mock(), Mock(), decorrelation)


@pytest.mark.parametrize(
    ("values", "error"),
    [
        param([1, np.inf], ValueError, id="infinity"),
        param([1, np.nan], ValueError, id="nan"),
        param([1, 1], ValueError, id="duplicates"),
        param([1], ValueError, id="only_one_value"),
        param([1, "a"], IterableValidationError, id="not_a_number"),
    ],
)
def test_invalid_values_numerical_discrete_parameter(values, error):
    """Providing invalid parameter values raises an exception."""
    with pytest.raises(error):
        NumericalDiscreteParameter(name="invalid_values", values=values)


@pytest.mark.parametrize(
    ("bounds", "error"),
    [
        param([-np.inf, np.inf], InfiniteIntervalError, id="infinite"),
        param([0, np.inf], InfiniteIntervalError, id="open_right"),
        param([-np.inf, 0], InfiniteIntervalError, id="open_left"),
        param([0, 0], ValueError, id="degenerate"),
    ],
)
def test_invalid_bounds_numerical_continuous_parameter(bounds, error):
    """Creating an unbounded parameter raises an exception."""
    with pytest.raises(error):
        NumericalContinuousParameter(name="invalid_values", bounds=bounds)


@pytest.mark.parametrize(
    ("values", "active_values", "error"),
    [
        param("ABC", None, ValueError, id="string"),
        param(["", "A"], None, ValueError, id="empty_string"),
        param(["A", "A"], None, ValueError, id="duplicates"),
        param(["A"], None, ValueError, id="only_one_value"),
        param(["A", 1], None, TypeError, id="not_a_string"),
        param(["A", "B"], [], ValueError, id="no_active_values"),
        param(["A", "B"], ["C"], ValueError, id="unknown_active_values"),
        param(["A", "B"], ["A", "A"], ValueError, id="duplicate_active_values"),
    ],
)
def test_invalid_values_categorical_parameter(values, active_values, error):
    """Providing invalid parameter values raises an exception."""
    with pytest.raises(error):
        CategoricalParameter(
            name="invalid_values", values=values, active_values=active_values
        )


def test_invalid_encoding_categorical_parameter():
    """Providing an invalid encoding raises an exception."""
    with pytest.raises(ValueError):
        CategoricalParameter(name="invalid_encoding", values=["A", "B"], encoding="enc")


@pytest.mark.parametrize(
    ("values", "active_values", "error"),
    [
        param("ABC", None, ValueError, id="string"),
        param(["", "A"], None, ValueError, id="empty_string"),
        param(["A", "A"], None, ValueError, id="duplicates"),
        param(["A"], None, ValueError, id="only_one_value"),
        param(["A", 1], None, TypeError, id="not_a_string"),
        param(["A", "B"], [], ValueError, id="no_active_values"),
        param(["A", "B"], ["C"], ValueError, id="unknown_active_values"),
        param(["A", "B"], ["A", "A"], ValueError, id="duplicate_active_values"),
    ],
)
def test_invalid_values_task_parameter(values, active_values, error):
    with pytest.raises(error):
        """Providing invalid (active) parameter values raises an exception."""
        TaskParameter(name="invalid_values", values=values, active_values=active_values)


@pytest.mark.skipif(
    not CHEM_INSTALLED, reason="Optional chem dependency not installed."
)
@pytest.mark.parametrize(
    ("data", "active_values", "error"),
    [
        param({"": "C", "A": "C"}, None, ValueError, id="empty_string"),
        param({"A": "C"}, None, ValueError, id="only_one_value"),
        param({"A": "C", 1: "C"}, None, TypeError, id="not_a_string"),
        param(
            {"A": "C", "B": "X", "C": "Y"}, None, ExceptionGroup, id="invalid_smiles"
        ),
        param(
            {"A": "CC", "B": "C-C", "C": "CCO", "D": "OCC"},
            None,
            ExceptionGroup,
            id="duplicate_substances",
        ),
        param({"A": "C", "B": "CC"}, [], ValueError, id="no_active_values"),
        param({"A": "C", "B": "CC"}, ["C"], ValueError, id="unknown_active_values"),
        param(
            {"A": "C", "B": "CC"}, ["A", "A"], ValueError, id="duplicate_active_values"
        ),
    ],
)
def test_invalid_data_substance_parameter(data, active_values, error):
    """Providing invalid substance data raises an exception."""
    with pytest.raises(error):
        SubstanceParameter(name="invalid_data", data=data, active_values=active_values)


@pytest.mark.skipif(
    not CHEM_INSTALLED, reason="Optional chem dependency not installed."
)
def test_invalid_encoding_substance_parameter():
    """Providing an invalid encoding raises an exception."""
    with pytest.raises(ValueError):
        SubstanceParameter(
            name="invalid_encoding", data={"A": "C", "B": "CC"}, encoding="enc"
        )


@pytest.mark.parametrize(
    ("data", "active_values"),
    [
        param(
            pd.DataFrame([[1, 2], [3, np.nan]], index=["A", "B"]),
            None,
            id="contains_nan",
        ),
        param(
            pd.DataFrame([[1, 2], [3, np.inf]], index=["A", "B"]),
            None,
            id="contains_inf",
        ),
        param(
            pd.DataFrame([[1, 2], [3, 4]], index=["A", "A"]), None, id="duplicate_idxs"
        ),
        param(pd.DataFrame([[1, 2]], index=["A"]), None, id="wrong_label_number"),
        param(
            pd.DataFrame([[1, 2], [1, 2]], index=["A", "B"]), None, id="zero_var_col"
        ),
        param(
            pd.DataFrame([[1, 2], [3, "a"]], index=["A", "B"]), None, id="wrong_type"
        ),
        param(
            pd.DataFrame([[1, 2], [3, 4]], index=["A", 1]), None, id="non_string_idx"
        ),
        param(
            pd.DataFrame([[1, 2], [3, 4]], index=["A", ""]), None, id="empty_string_idx"
        ),
        param(
            pd.DataFrame([[1, 2], [1, 2], [3, 4]], index=["A", "B", "C"]),
            None,
            id="duplicate_rows",
        ),
        param(
            pd.DataFrame([[1, 2], [3, 4]], index=["A", "B"]),
            [],
            id="no_active_values",
        ),
        param(
            pd.DataFrame([[1, 2], [3, 4]], index=["A", "B"]),
            ["C"],
            id="unknown_active_values",
        ),
        param(
            pd.DataFrame([[1, 2], [3, 4]], index=["A", "B"]),
            ["A", "A"],
            id="duplicate_active_values",
        ),
    ],
)
def test_invalid_data_custom_parameter(data, active_values):
    """Providing an invalid custom encoding raises an exception."""
    with pytest.raises(ValueError):
        CustomDiscreteParameter(
            name="invalid_data", data=data, active_values=active_values
        )



"""Hypothesis strategies for objectives."""

import hypothesis.strategies as st

from baybe.objectives.desirability import DesirabilityObjective
from baybe.objectives.enum import Scalarizer
from baybe.objectives.pareto import ParetoObjective
from baybe.objectives.single import SingleTargetObjective
from baybe.targets import NumericalTarget, TargetMode
from baybe.targets.numerical import _VALID_TRANSFORMATIONS

from ..hypothesis_strategies.basic import finite_floats
from ..hypothesis_strategies.targets import numerical_targets
from ..hypothesis_strategies.utils import intervals as st_intervals

_intervals = st_intervals(exclude_fully_unbounded=True, exclude_half_bounded=True)

_target_lists = st.lists(
    numerical_targets(_intervals), min_size=2, unique_by=lambda t: t.name
)


def single_target_objectives():
    """Generate :class:`baybe.objectives.single.SingleTargetObjective`."""
    return st.builds(SingleTargetObjective, target=numerical_targets())


@st.composite
def desirability_objectives(draw: st.DrawFn):
    """Generate :class:`baybe.objectives.desirability.DesirabilityObjective`."""
    targets = draw(_target_lists)
    weights = draw(
        st.lists(
            finite_floats(min_value=0.0, exclude_min=True),
            min_size=len(targets),
            max_size=len(targets),
        )
    )
    scalarizer = draw(st.sampled_from(Scalarizer))
    return DesirabilityObjective(targets, weights, scalarizer)


@st.composite
def _pareto_targets(draw: st.DrawFn):
    """Generate :class:`baybe.targets.numerical.NumericalTarget`."""
    name = draw(st.text(min_size=1))
    mode = draw(st.sampled_from(TargetMode))

    if mode is TargetMode.MATCH:
        transformation = draw(st.sampled_from(_VALID_TRANSFORMATIONS[mode]))
        bounds = draw(
            st_intervals(exclude_half_bounded=True, exclude_fully_unbounded=True)
        )
    else:
        transformation = None
        bounds = None

    return NumericalTarget(
        name=name, mode=mode, bounds=bounds, transformation=transformation
    )


_pareto_target_lists = st.lists(
    _pareto_targets(), min_size=2, unique_by=lambda t: t.name
)


@st.composite
def pareto_objectives(draw: st.DrawFn):
    """Generate :class:`baybe.objectives.pareto.ParetoObjective`."""
    return ParetoObjective(draw(_pareto_target_lists))



"""Module for persisting benchmarking results."""

from benchmarks.persistence.persistence import (
    LocalFileObjectStorage,
    PathConstructor,
    S3ObjectStorage,
)

__all__ = ["PathConstructor", "S3ObjectStorage", "LocalFileObjectStorage"]
